(herald pp (env tsys))

;;;; a pretty-printer

(define-operation (pretty-print obj port)
  (print obj port))

;;; handler for lists.

(define (pp-list x port)
  (cond ((read-macro-inverse x)
         => (lambda (inverse)
              (writes port inverse)
              (pretty-print (cadr x) port)))
        ((print-width-greater? x (fx- (line-length port) (hpos port)))
         ;; if ordinary print won't win...
         (pp-list-vertically x port))
        (else
         (pp-list-horizontally x port))))

;++ obsolete

(define (print-width-greater? obj n)
  (catch abort
    (let ((count 0))
      (print obj (object nil
                   ((writec self char)
                    (ignore char)
                    (set count (fx+ count 1))
                    (if (fx> count n) (abort t)))
                   ((writes self string)
                    (set count (fx+ count (string-length string)))
                    (if (fx> count n) (abort t)))
                   ((output-port? self) '#t)
                   ((port? self) '#t)
                   ((print-type-string self) "Output-port"))))
    nil))


(define (pp-list-vertically   x port)
  (maybe-pp-list-vertically t x port))

(define (pp-list-horizontally x port)
  (maybe-pp-list-vertically nil x port))

(define (too-wide? x port)
  (print-width-greater? x (fx- (line-length port) (hpos port))))

(define (maybe-pp-list-vertically vertical? list port)
  (writec port #\()
  (if (null? list)
      (writec port #\))
      (let ((oldest-hpos (hpos port))
	    (first-too-wide? (too-wide? (car list) port)))
	(pretty-print (if (syntax-descriptor? (car list))
			  (identification (car list))
			  (car list))
		      port)
	(if (and vertical?
		 (not (never-break? (car list)))
		 (not (null? (cdr list)))
		 (or first-too-wide?
		     (pair? (car list))
		     (too-wide? (cadr list) port)))
	    (indent-newline (fx- oldest-hpos 1) port))
	(let ((old-hpos (cond ((pair? (car list)) oldest-hpos)
			      ((always-indent-two? (car list))
			       (fx+ oldest-hpos 1))
			      (else (fx+ (hpos port) 1)))))
	  (iterate tail ((indent? nil) (l (cdr list)))
	    (if (pair? l) 
		(block (if indent?
			   (indent-newline old-hpos port)
			   (writec port #\space)) ; not (space port)!
		       (let ((prev-too-wide? (too-wide? (car l) port)))
			 (pretty-print (car l) port)
			 (tail (and vertical?
				    (or (fx= old-hpos oldest-hpos)
					(fx= old-hpos (fx+ oldest-hpos 1))
					(never-break? (car list))
					prev-too-wide?
					(too-wide? (cadr l) port)))
			       (cdr l))))
		(block (if (not (null? l))
			   (block (format port " . ")
				  (if indent? (indent-newline old-hpos port))
				  (pretty-print l port)))
		       (writec port #\)))))))))

(define (always-indent-two? x)
  (and (symbol? x)
       (let ((str (symbol->string x)))
	 (or (string-equal? "DEF" (substring str 0 3))
	     (string-equal? "LET" (substring str 0 3))
	     (string-equal? "LAMBDA" str)))))

(define (never-break? x)
  (or (always-indent-two? x)
      (and (symbol? x)
	   (let ((str (symbol->string x)))
	     (or (string-equal? "IF" str)
		 (string-equal? "COND" str))))))

(define (indent-newline x port)
  (newline port)
  (let ((x (if (>= x (line-length port))
	       (fx- (line-length port) 1)
	       x)))
    (iterate loop ((i 0))
      (cond ((fx>= i x) (no-value))
	    (else
	     (writec port #\space)
	     (loop (fx+ i 1)))))))


;;; find printed representation for internal representation of read
;;; macro.

(define (read-macro-inverse x)
  (cond ((and (pair? x)
              (pair? (cdr x))
              (null? (cddr x)))
         (case (car x)
           ((quote)            "'")
           ((quasiquote)       "`")
           ((unquote)          ",")
           ((unquote-splicing) ",@")
           (else nil)))
        (else nil)))

;;; "user interface" stuff

(define (*pp-symbol symbol env)
  (*pp (cond ((syntax-table-entry (env-syntax-table env) symbol)
              => identity)
             ((env-lookup env symbol nil nil)
              => (lambda (loc)
                   (let ((val (contents loc)))
                     (cond ((nonvalue? val) 'unbound)
                           (else val)))))
             (else 'unbound))))

(define (*pp obj)
  (let ((obj (or (disclose obj) obj)))
    (let ((port (terminal-output)))
      (fresh-line port)
      (cond ((and (procedure? obj) (where-defined obj))
             => (lambda (where) (format port "~&;see ~a~%" where)))
            (else
             (pretty-print obj port)))
      (fresh-line port)
      repl-wont-print)))
